import string
from itertools import zip_longest
from pathlib import Path
from textwrap import dedent
from typing import List

import typer
from fyeah import f

from nef_pipelines.lib.isotope_lib import ATOM_TO_ISOTOPE
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.sequence_lib import MoleculeTypes, sequence_from_entry
from nef_pipelines.lib.structures import NewPeak
from nef_pipelines.lib.translation_lib import translate_new_peak
from nef_pipelines.lib.util import STDIN, exit_error, parse_comma_separated_options
from nef_pipelines.transcoders.sparky import import_app
from nef_pipelines.transcoders.sparky.importers.shifts import (
    _exit_if_number_chain_codes_and_file_names_dont_match,
)
from nef_pipelines.transcoders.sparky.sparky_lib import parse_peaks


# TODO: this needs to be moved to a library
class SparkyPeakListException(Exception):
    pass


class IncompatibleDimensionTypesException(SparkyPeakListException):
    pass


app = typer.Typer()

DEFAULT_NUCLEI_HELP = (
    "nuclei to use for each dimension, if not defined they are guessed from the assignments"
    "or an error is reported"
)


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    chain_codes: List[str] = typer.Option(
        None,
        "--chains",
        help="chain codes as a list of names separated by commas, repeated calls will add further chains [default A] "
        "one per file. If only one chain code is supplied it applies to all files",
        metavar="<CHAIN-CODES>",
    ),
    frame_name: str = typer.Option(
        "sparky_{file_name}",
        "-f",
        "--frame-name",
        help="a templated name for the frame {file_name} will be replaced by input filename",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type peaks.txt", metavar="<SPARKY-peaks>.txt"
    ),
    nuclei: List[str] = typer.Option([], help=DEFAULT_NUCLEI_HELP),
    default_chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain-code",
        help="default chain code to use if none is provided in the file",
    ),
    molecule_type: MoleculeTypes = typer.Option(
        MoleculeTypes.PROTEIN, help="the type of molecule"
    ),
    no_validate: bool = typer.Option(
        False,
        help="if set don't validate the peaks against the input sequence if provided",
    ),
    spectrometer_frequency: float = typer.Option(
        600.123456789, help="spectrometer frequency in MHz"
    ),
):
    """convert sparky peaks file <SPARKY-PEAKS>.txt to NEF"""

    chain_codes = parse_comma_separated_options(chain_codes)

    if not chain_codes:
        chain_codes = ["A"]

    if len(chain_codes) == 1 and len(file_names) > 1:
        chain_codes = chain_codes * len(file_names)

    # make this a library function
    _exit_if_number_chain_codes_and_file_names_dont_match(chain_codes, file_names)

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    sequence = sequence_from_entry(entry) if not no_validate else None

    nuclei = parse_comma_separated_options(nuclei)

    file_names_and_lines = {}
    for file_name, chain_code in zip_longest(file_names, chain_codes, fillvalue=None):

        if file_name is None:
            continue

        if chain_code is None:
            chain_code = chain_codes[-1]

        try:
            with open(file_name, "r") as fp:
                lines = fp.readlines()
        except IOError as e:
            msg = f"""
                    while reading sparky peaks file {file_name} there was an error reading the file
                    the error was: {e}
                """
            exit_error(msg, e)

        file_names_and_lines[file_name] = lines

    entry = pipe(
        entry,
        frame_name,
        file_names_and_lines,
        chain_code,
        sequence,
        input_dimensions=nuclei,
        spectrometer_frequency=spectrometer_frequency,
        molecule_type=molecule_type,
    )

    print(entry)


def pipe(
    entry,
    frame_code_template,
    file_names_and_lines,
    chain_code,
    sequence,
    input_dimensions,
    spectrometer_frequency,
    molecule_type=MoleculeTypes.PROTEIN,
):

    sparky_frames = []

    for file_name, lines in file_names_and_lines.items():

        sparky_peaks = parse_peaks(
            lines,
            file_name=file_name,
            molecule_type=molecule_type,
            chain_code=chain_code,
            sequence=sequence,
        )

        sparky_peaks = [translate_new_peak(peak) for peak in sparky_peaks]

        dimensions = _guess_dimensions_if_not_defined_or_throw(
            sparky_peaks, input_dimensions
        )

        dimensions = [{"axis_code": dimension} for dimension in dimensions]

        file_name = Path(file_name).stem  # used in f method...

        frame_code = f(frame_code_template)

        frame = peaks_to_frame(
            sparky_peaks, dimensions, spectrometer_frequency, frame_code=frame_code
        )

        sparky_frames.append(frame)

    return add_frames_to_entry(entry, sparky_frames)


def _guess_dimensions_if_not_defined_or_throw(
    peaks: List[NewPeak], input_dimensions: List[str]
) -> List[str]:
    results_by_dimension = {
        i: dimension for i, dimension in enumerate(input_dimensions)
    }

    guessed_dimensions = {}

    for peak_number, peak in enumerate(peaks, start=1):

        for dimension_index, shift in enumerate(peak.shifts):
            atom_name = shift.atom.atom_name

            if len(atom_name) == 0:
                atom_name = None

            atom_type = None
            if atom_name != UNUSED and atom_name is not None:
                atom_type = atom_name.strip()[0]

            if atom_type is not None and atom_type.lower() not in string.ascii_letters:
                atom_type = None

            guessed_dimension_isotopes = guessed_dimensions.setdefault(
                dimension_index, set()
            )
            if atom_type and atom_type in ATOM_TO_ISOTOPE:
                isotope_code = ATOM_TO_ISOTOPE[atom_type]
                guessed_dimension_isotopes.add(isotope_code)
            else:
                guessed_dimension_isotopes.add(UNUSED)

    for guessed_dimension_index in guessed_dimensions:
        if guessed_dimension_index not in results_by_dimension:

            guessed_dimension = guessed_dimensions[guessed_dimension_index]
            if len(guessed_dimension) > 1 and UNUSED in guessed_dimension:
                guessed_dimension.remove(UNUSED)

            if len(guessed_dimensions[guessed_dimension_index]) > 1:

                _raise_multiple_guessed_isotopes(
                    guessed_dimensions, guessed_dimension_index
                )

    for dimension_index in guessed_dimensions:
        if dimension_index not in results_by_dimension:
            results_by_dimension[dimension_index] = list(
                guessed_dimensions[dimension_index]
            )[0]

    max_dimension = max(results_by_dimension.keys())

    result = []
    for i in range(max_dimension + 1):
        result.append(results_by_dimension[i])

    return result


def _raise_multiple_guessed_isotopes(guessed_dimensions, guessed_dimension_index):
    dim_info = []
    for dimension_index in sorted(guessed_dimensions).keys():
        dim_data = f"{dimension_index}: {' '.join(guessed_dimensions.values())}"
        dim_info.append(dim_data)
    msg = f"""\
                    no isotope provided for dimension {guessed_dimension_index + 1} and multiple possibilities
                    found in atom names, please use explicit isotope codes using dimension-nuclei

                    deduced codes for dimensions:


                """
    msg = dedent(msg)
    msg += dim_info
    raise IncompatibleDimensionTypesException(msg)



from textwrap import dedent

import typer
from freezegun import freeze_time
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import assert_lines_match
from nef_pipelines.lib.util import get_version
from nef_pipelines.main import create_nef_app
from nef_pipelines.tools.header import header

create_nef_app()


app = typer.Typer()
app.command()(header)

runner = CliRunner()

EXPECTED_TEMPLATE = f"""
    data_%(name)s

    save_nef_nmr_meta_data
       _nef_nmr_meta_data.sf_category      nef_nmr_meta_data
       _nef_nmr_meta_data.sf_framecode     nef_nmr_meta_data
       _nef_nmr_meta_data.format_name      nmr_exchange_format
       _nef_nmr_meta_data.format_version   1.1
       _nef_nmr_meta_data.program_name     NEFPipelines
       _nef_nmr_meta_data.script_name      header.py
       _nef_nmr_meta_data.program_version  {get_version()}
       _nef_nmr_meta_data.creation_date    2012-01-14T12:00:01.123456
       _nef_nmr_meta_data.uuid             NEFPipelines-2012-01-14T12:00:01.123456-1043321819

       loop_
          _nef_run_history.run_number
          _nef_run_history.program_name
          _nef_run_history.program_version
          _nef_run_history.script_name


       stop_

    save_"""


def get_expected(name):
    expected = EXPECTED_TEMPLATE % {"name": name}
    expected = dedent("\n".join(expected.split("\n")[1:]))
    return expected


# noinspection PyUnusedLocal
@freeze_time("2012-01-14 12:00:01.123456")
def test_nef_default_header(fixed_seed):

    result = runner.invoke(app, [])

    assert result.exit_code == 0

    assert_lines_match(get_expected("nef"), result.stdout)


# noinspection PyUnusedLocal
@freeze_time("2012-01-14 12:00:01.123456")
def test_nef_named_header(fixed_seed):

    result = runner.invoke(app, ["test"])
    assert result.exit_code == 0

    assert_lines_match(get_expected("test"), result.stdout)



from pathlib import Path

import typer

from nef_pipelines import nef_app

if nef_app:
    # noinspection PyUnusedLocal
    @nef_app.app.command()
    def stream(
        file_name: Path = typer.Argument(
            ..., help="nef file to stream [- indicates stdin]", metavar="<nef file>"
        )
    ):
        "- stream a nef file"
        with open(file_name) as file_h:
            for line in file_h:
                print(line, end="")



from collections.abc import MutableSet
from dataclasses import dataclass
from typing import Optional, Union


class IdentitySet(MutableSet):
    """
    A set that stores object by there identity
    """

    key = id  # should return a hashable object

    def __init__(self, iterable=()):
        self.map = {}  # id -> object
        self |= iterable  # add elements from iterable to the set (union)

    def __len__(self):  # Sized
        return len(self.map)

    def __iter__(self):  # Iterable
        return iter(self.map.values())

    def __contains__(self, x):  # Container
        return self.key(x) in self.map

    def add(self, value):  # MutableSet
        """Add an element."""
        self.map[self.key(value)] = value

    def discard(self, value):  # MutableSet
        """Remove an element.  Do not raise an exception if absent."""
        self.map.pop(self.key(value), None)

    def update(self, iterable):
        for elem in iterable:
            self.add(elem)

    def __repr__(self):
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))


def return_path_and_value(iterable, item):
    return iter.get_path(item), item.value


def return_entity(iter, item):
    return item


def item_only(iter, item):
    return item


def object_children(target, parent):
    result = ()

    if hasattr(target, "model_fields"):
        result = tuple(
            [
                ObjectIter.Entry(target, field_name, getattr(target, field_name))
                for field_name in target.model_fields.keys()
            ]
        )

    elif isinstance(target, list):
        result = tuple(
            [ObjectIter.Entry(target, i, elem) for i, elem in enumerate(target)]
        )

    elif isinstance(target, dict):
        result = tuple(
            [ObjectIter.Entry(target, key, value) for key, value in target.items()]
        )

    return result


class ObjectIter:
    @dataclass
    class Entry:
        target: object
        selector: Optional[Union[int, str]]
        value: object
        is_leaf: bool = False

        def is_root(self):
            return self.target is None and self.selector is None and self.value is None

    def __init__(
        self, child_lister=object_children, result_generator=return_path_and_value
    ):
        self._child_lister = child_lister
        self._result_generator = result_generator
        self._child_parent = ObjectIter.Parents()

    class Parents:
        def __init__(self):
            self._parents = {}

        def add(self, child_info, parent):

            self._parents[id(child_info)] = parent

        def add_all(self, children_info, parent):
            for child_info in children_info:
                self.add(child_info, parent)

        def parent(self, child):
            child_id = id(child)
            return self._parents[child_id] if child_id in self._parents.keys() else None

        def get_path(self, child):

            result = []
            target = child

            while not target.is_root():
                result.append(target.selector)
                target = self.parent(target)

            return tuple(reversed(result[1:]))

    def get_path(self, item):
        return self._child_parent.get_path(item)

    def get_parent(self, item):

        return self._child_parent.parent(item)

    def iter(self, target):

        root_entry = ObjectIter.Entry(None, None, None)
        entry = ObjectIter.Entry(None, None, target)
        self._child_parent.add(entry, root_entry)

        yield self._result_generator(self, entry)

        visited = IdentitySet()
        non_visited = IdentitySet()

        children = self._child_lister(target, root_entry)
        self._child_parent.add_all(children, entry)

        non_visited.update(children)

        while non_visited:
            item = non_visited.pop()

            if item in visited:
                continue

            visited.add(item)

            yield self._result_generator(self, item)

            children = [
                elem
                for elem in self._child_lister(item.value, item)
                if not elem.is_leaf
            ]
            self._child_parent.add_all(children, item)

            non_visited.update(children)



# original implementation by esther
import sys
from dataclasses import dataclass
from string import digits
from typing import List, Union

import typer
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    read_entry_from_stdin_or_exit,
    select_frames_by_name,
)
from nef_pipelines.lib.shift_lib import nef_frames_to_shifts
from nef_pipelines.lib.util import exit_error, is_float, is_int
from nef_pipelines.transcoders.mars import export_app

STDOUT_PATH = "-"

app = typer.Typer()

# TODO: name translations
# TODO: correct weights
# TODO: move utuilities to lib
# TODO: support multiple chains

REMOVE_DIGITS = str.maketrans("", "", digits)


def has_numbers(input: str) -> bool:
    return any(char.isdigit() for char in input)


# noinspection PyUnusedLocal
@export_app.command()
def shifts(
    shift_frames: List[str] = typer.Option(
        [],
        "-f",
        "--frame",
        help="selector for the shift restraint frames to use, can be called multiple times and include wild cards",
    ),
    chain: str = typer.Option(
        [],
        "-c",
        "--chain",
        help="chains to export, to add multiple chains use repeated calls  [default: 'A']",
        metavar="<CHAIN-CODE>",
    ),
    output_file: str = typer.Argument(
        None,
        help="file name to output to [default <entry_id>_shifts.tab] for stdout use -",
        metavar="<MARS_SHIFT_FILE>",
    ),
):
    """- write a mars chemical shift file"""

    if len(shift_frames) == 0:
        shift_frames = ["*"]

    entry = read_entry_from_stdin_or_exit()

    output_file = f"{entry.entry_id}_shifts.tab" if output_file is None else output_file

    frames = entry.get_saveframes_by_category("nef_chemical_shift_list")

    frames = select_frames_by_name(frames, shift_frames)

    if len(frames) == 0:
        exit_error("no shift frames selected")

    shifts = nef_frames_to_shifts(frames)

    @dataclass(frozen=True, order=True)
    class PseudoAtom:
        sequence_code: Union[int, str]
        residue_name: str
        negative_offset: int
        atom_name: str

    pseudo_atom_shifts = {}
    for shift in shifts:

        # first deal with completely unassigned residues chain_code @- and sequence_code @xxxx where xxx
        # is the pseudo residue
        chain_code = str(shift.atom.residue.chain_code)
        sequence_code = str(shift.atom.residue.sequence_code)
        atom_name = shift.atom.atom_name
        if (
            chain_code.startswith("@") or chain_code.startswith("#")
        ) and sequence_code.startswith("@"):
            sequence_code = sequence_code.lstrip("@")
            sequence_code_fields = sequence_code.split("-")

            if len(sequence_code_fields) > 2:
                continue

            sequence_code = sequence_code_fields[0]
            if not is_int(sequence_code):
                continue
            sequence_code = int(sequence_code)

            negative_offset = (
                sequence_code_fields[1] if len(sequence_code_fields) == 2 else 0
            )

            if not is_int(negative_offset):
                continue
            else:
                negative_offset = int(negative_offset)

            if negative_offset not in (0, 1):
                continue

            atom_name = atom_name.replace("@", "")
            if has_numbers(atom_name):
                continue

            pseudo_atom = PseudoAtom(
                sequence_code,
                shift.atom.residue.residue_name,
                negative_offset,
                atom_name=atom_name,
            )

            if not is_float(shift.value):
                continue

                # TODO: should be value
            value = float(shift.value)

            pseudo_atom_shifts[pseudo_atom] = value

    headings = (
        ("H", ("H", 0)),
        ("N", ("N", 0)),
        ("Ca", ("CA", 0)),
        ("Ca-1", ("CA", 1)),
        ("Cb", ("CB", 0)),
        ("Cb-1", ("CB", 1)),
        ("CO", ("C", 0)),
        ("CO-1", ("C", 1)),
        ("HA", ("HA", 0)),
        ("HA-1", ("HA", 1)),
    )
    base_headers = (
        "",
        "H",
        "N",
        "CA",
        "CA-1",
        "CB",
        "CB-1",
        "CO",
        "CO-1",
        "HA",
        "HA-1",
    )

    pseudo_residues = {}
    for pseudo_atom in pseudo_atom_shifts:
        key = (pseudo_atom.atom_name, pseudo_atom.negative_offset)
        pseudo_residues.setdefault(pseudo_atom.sequence_code, {})[key] = pseudo_atom

    headers = _filter_headings_by_pseudoatoms(base_headers, pseudo_residues)

    lines = []
    for pseudo_residue_num in sorted(pseudo_residues):
        pseudo_atoms = pseudo_residues[pseudo_residue_num]

        line = []
        lines.append(line)

        line.append("PR_" + str(pseudo_residue_num))

        for heading, heading_key in headings:

            if heading_key in pseudo_atoms:

                pseudo_atom = pseudo_atoms[heading_key]
                shift = pseudo_atom_shifts[pseudo_atom]

                line.append("%-7.3f    " % shift)
            elif heading.upper() in headers:
                line.append("-         ")

    file_h = sys.stdout if output_file == "-" else open(output_file, "w")

    print(tabulate(lines, headers=headers, tablefmt="plain"), file=file_h)

    if output_file != STDOUT_PATH:
        file_h.close()

        if not sys.stdout.isatty():
            print(entry)


def _filter_headings_by_pseudoatoms(base_headers, pseudo_residues):
    atom_names = set()
    for pseudo_residue in pseudo_residues.values():
        for pseudo_atom in pseudo_residue.values():
            offset = (
                f"-{pseudo_atom.negative_offset}"
                if pseudo_atom.negative_offset == 1
                else ""
            )
            atom_names.add(f"{pseudo_atom.atom_name.upper()}{offset}")
    headers = []
    for header in base_headers:
        if header == "" or header in atom_names:
            headers.append(header)
    return headers



from textwrap import dedent
from typing import Dict, Iterable, List, Tuple

from pyparsing import (
    Forward,
    Group,
    ParseException,
    ParserElement,
    ParseResults,
    Suppress,
    Word,
    ZeroOrMore,
    printables,
    restOfLine,
)

from nef_pipelines.lib.nef_lib import UNUSED
from nef_pipelines.lib.sequence_lib import get_residue_name_from_lookup
from nef_pipelines.lib.structures import (
    AtomLabel,
    SequenceResidue,
    ShiftData,
    ShiftList,
)
from nef_pipelines.lib.util import exit_error


def _process_emptys_and_singles(value: ParseResults) -> ParseResults:
    """
    a parse action for tcl_parser that ignores empty lists and promotes lists with a single item to a single item
    Args:
        value (ParseResults): parse resultsd to be modified

    Returns:
        ParseResults: corrected parse results
    """

    for i, item in enumerate(value):
        if len(item) == 0:
            value[i] = ""

    if len(value) == 1:
        value = value[0]

    return value


# @functools.cache
def get_tcl_parser() -> ParserElement:
    """
    build a simple tcl parser suitable for nmrview files and cach it

    Returns:
        pyparsing parser

    """

    # TODO this should be printables  excluding : " {  }
    simple_word = Word(initChars=printables, excludeChars='"{}')
    simple_word.setName("simple_word")

    expression = Forward()
    expression.setName("expression")

    DBL_QUOTE = Suppress('"')
    LEFT_PAREN = Suppress("{")
    RIGHT_PAREN = Suppress("}")

    quoted_simple_word = DBL_QUOTE + simple_word + DBL_QUOTE
    quoted_simple_word.setName("quoted_simple_word")

    quoted_complex_word = Group(DBL_QUOTE + ZeroOrMore(expression) + DBL_QUOTE)
    quoted_complex_word.setName("quoted complex word")

    complex_list = Group(LEFT_PAREN + ZeroOrMore(expression) + RIGHT_PAREN)
    complex_list.setName("complex list")

    expression << (
        simple_word | quoted_simple_word | quoted_complex_word | complex_list
    )

    remainder = restOfLine()
    remainder.setName("remainder")

    top_level = ZeroOrMore(expression)
    top_level.setParseAction(_process_emptys_and_singles)
    top_level.setName("phrase") + restOfLine()

    return top_level


def parse_tcl(in_str, file_name="unknown", line_no=0) -> ParseResults:
    """
    parse a tcl data file or fragment using pyparsing

    Args:
        in_str (str):  tcl source
        file_name (str):  file name for error reporting
        line_no (str):  base line number for error reporting, the line no reported by py parsing will be added to this

    Returns:
        ParseResults: pyparsing parse result
    """

    parser = get_tcl_parser()

    result = None
    try:
        result = parser.parseString(in_str, parseAll=True)
    except ParseException as pe:
        line_no += pe.lineno
        msg = f"""\
                    Failed while parsing tcl at line: {line_no} in file: {file_name}

                    Explanation:
                """

        msg = dedent(msg)

        msg += pe.explain_exception(pe)

        exit_error(msg)

    return result


# TODO: add line info for better error handling
def parse_float_list(line: str, line_no: int) -> List[float]:
    """
     parse a tcl list into a list of floats using the tcl parser

    Args:
        line (str): the input string
        line_no: line number for error reporting

    Returns:
       List[float]: a list of floats
    """

    raw_fields = []
    parsed_tcl = parse_tcl(line)
    for field in parsed_tcl:
        if isinstance(field, ParseResults):
            raw_fields.extend(field)
        elif isinstance(field, str):
            raw_fields.append(field)
        else:
            exit_error(
                f"Error: unexpected internal error tcl failed to parse: '{line}' was not parsed properly got: \
                {parsed_tcl} field type was: {field.__class__}"
            )

    result = []
    for field_index, field in enumerate(raw_fields):
        try:
            field = float(field)
        except ValueError as e:
            msg = f"Couldn't convert sweep width {field_index} [{field}] to float for line {line} at line number \
                  {line_no}"
            exit_error(msg, e)
        result.append(field)

    return result


def read_sequence(
    sequence_lines: Iterable[str],
    chain_code: str = "A",
    sequence_file_name: str = "unknown",
) -> List[SequenceResidue]:

    """
    read an nmrview sequence from a file

    Args:
        sequence_lines (Iterable[str]): the lines for the file
        chain_code (str): a chaning code to use
        sequence_file_name (str): the file being read from for reporting errors

    Returns:
        List[SequenceResidue]: The residues as structures
    """

    start_residue = 1
    result = []
    for i, line in enumerate(sequence_lines):
        line = line.strip()
        fields = line.split()

        msg = f"""nmview sequences have one residue name per line,
                  except for the first line which can also contain a starting residue number,
                  at line {i + 1} i got {line} in file {sequence_file_name}
                  line was: {line}"""

        if len(fields) > 1 and i != 0:
            exit_error(msg)

        if i == 0 and len(fields) > 2:
            exit_error(
                f"""at the first line the should be one 3 letter code and an optional residue number
                           in file {sequence_file_name} at line {i+1} got {len(fields)} fields
                           line was: {line}"""
            )

        if i == 0 and len(fields) == 2:
            try:
                start_residue = int(fields[1])
            except ValueError:
                msg = f"""couldn't convert second field {fields[0]} to an integer
                          at line {i + 1} in file {sequence_file_name}
                          line was: {line}
                        """
                exit_error(msg)

        if len(fields) > 0:
            result.append(SequenceResidue(chain_code, start_residue + i, fields[0]))

    return result


def parse_shifts(
    lines: Iterable[str],
    chain_seqid_to_type: Dict[Tuple[str, int], str],
    chain_code: str = "A",
    file_name="unknown",
) -> ShiftList:

    shifts = []
    for i, line in enumerate(lines):

        line = line.strip()

        if len(line) == 0:
            continue

        fields = line.split()
        num_fields = len(fields)
        if num_fields != 3:
            msg = f"""An nmrview ppm.out file should have 3 fields per line
                    i got {num_fields} at line {i + 1}
                    with data: {line}"""
            exit_error(msg)

        shift, stereo_specificty_code = fields[1:]

        atom_fields = fields[0].split(".")
        num_atom_fields = len(atom_fields)
        if num_atom_fields != 2:
            msg = f"""An nmrview ppm.out file should have atom specfiers of the form '1.CA'
                                        i got{num_atom_fields} at line {i + 1}
                                        with data: {line}"""
            exit_error(msg)
        residue_code, atom = atom_fields

        try:
            residue_code = int(residue_code)
        except ValueError:
            msg = f"""An nmrview residue number should be an integer
                      i got {residue_code} at line {i + 1}"""
            exit_error(msg)

        try:
            shift = float(shift)
        except ValueError:
            msg = f"""A chemical shift should be a float
                      i got {shift} at line {i + 1}"""
            exit_error(msg)

        try:
            stereo_specificty_code = int(stereo_specificty_code)
        except ValueError:
            msg = f"""An nmrview stereo specificty code should be an integer
                      i got {stereo_specificty_code} at line {i + 1}"""
            exit_error(msg)

        if stereo_specificty_code not in [1, 2, 3]:
            msg = f"""An nmrview stereo specificty code should be an integer between 1 and 3,
                      i got {stereo_specificty_code} at line {i + 1}"""
            exit_error(msg)

        residue_name = get_residue_name_from_lookup(
            chain_code, residue_code, chain_seqid_to_type
        )

        if (residue_code != UNUSED and chain_code != UNUSED) and residue_name == UNUSED:
            msg = f"""\
            residue not defined for chain {chain_code} residue {residue_code}
            line number {i+1}
            line was |{line}|
            in file {file_name}
            did you read a sequence?"""
            msg = dedent(msg)
            exit_error(msg)

        atom = AtomLabel(SequenceResidue(chain_code, residue_code, residue_name), atom)
        shift = ShiftData(atom, shift)
        shifts.append(shift)

    result = ShiftList(shifts)

    return result
